from typing import Unpack, NotRequired, Sequence
from chess import STARTING_FEN
from haskellian import AsyncIter, Iter
from chess_utils import position_idx
from .beam import predict_async as beam_predict_async, predict_sync as beam_predict_sync, SearchParams, FENlessSearchParams, Child
from .util import prefetched_logprobs, PredictFn, Annotations, logprob
from .types import AutoPred

class PredictParams(SearchParams):
  batch_size: NotRequired[int]
  prefetch: NotRequired[int]


def predict(
  predict: PredictFn, annotations: list[Annotations] | None = None,
  **params: Unpack[PredictParams]
) -> AsyncIter[Sequence[AutoPred]]:
  return predict_async(predict, annotations, **params) \
    .map(lambda nodes: list(map(AutoPred.of, nodes)))

def predict_async(
  predict: PredictFn, annotations: list[Annotations] | None = None, *,
  fen: str | None = None, batch_size: int = 8, prefetch: int = 2,
  **params: Unpack[FENlessSearchParams]
) -> AsyncIter[list[Child]]:
  """Beam decoding across the forest of moves stemming from `start_fens`
  - Yields predictions as the beams converge (i.e. agree on a single move) or the search stops (because no legal moves have high enough probability)
    - Thus, a bigger `beam_width` can increase accuracy but also prediction time by more than a constant factor
  - `beam_width` defaults to `4`
  """
  fen = fen or STARTING_FEN
  start_ply = position_idx(fen)
  return beam_predict_async(
    prefetched_logprobs(predict, annotations, start_ply=start_ply, batch_size=batch_size, prefetch=prefetch),
    fen=fen, **params
  )

def predict_sync(
  preds: list[list[list[tuple[str, float]]]], annotations: list[Annotations] | None = None,
  **params: Unpack[PredictParams]
) -> Iter[Sequence[AutoPred]]:
  return predict_sync_raw(preds, annotations, **params) \
    .map(lambda nodes: list(map(AutoPred.of, nodes)))

def predict_sync_raw(
  preds: list[list[list[tuple[str, float]]]],
  annotations: list[Annotations] | None = None, *,
  fen: str | None = None, **params: Unpack[FENlessSearchParams]
) -> Iter[list[Child]]:
  """Beam decoding across the forest of moves stemming from `start_fens`
  - Yields predictions as the beams converge (i.e. agree on a single move) or the search stops (because no legal moves have high enough probability)
    - Thus, a bigger `beam_width` can increase accuracy but also prediction time by more than a constant factor
  - `beam_width` defaults to `4`
  """
  fen = fen or STARTING_FEN
  start_ply = position_idx(fen)
  logprobs = (logprob(p, annotations) for p in preds[start_ply:])
  return beam_predict_sync(logprobs, fen=fen, **params)


from typing import Protocol, Iterable, TypedDict, NotRequired
from dataclasses import dataclass
import numpy as np
import chess
from chess_utils import CapturablePiece, unchecked_san, captured_piece, fen_after

class Logprob(Protocol):
  def __call__(self, san: str, captured_piece: CapturablePiece | None, /) -> float:
    ...
    
class AggregateLogps(Protocol):
  def __call__(self, logp_ocr: float, logp_prior: float, /) -> float:
    ...
    
@dataclass
class Node:
  fen: str
  sum_logp: float
  
  def __init__(self, fen: str, sum_logp = None):
    self.fen = fen
    self.sum_logp = sum_logp or 0
  
@dataclass
class Child(Node):
  sum_logp: float
  fen: str
  logp: float
  san: str
  parent: Node
  
  def __init__(self, fen: str, logp: float, san: str, parent: Node, sum_logp = None):
    self.fen = fen
    self.sum_logp = sum_logp or logp + parent.sum_logp
    self.logp = logp
    self.san = san
    self.parent = parent
    
class SuccParams(TypedDict):
  logp_min: NotRequired[float]
  logp_ocr_min: NotRequired[float]

def successors(
  node: Node, uci_priors: dict[str, float], lp: Logprob, agg_lp: AggregateLogps, *,
  logp_min: float = np.log(0.02), logp_ocr_min: float = np.log(0.02),
) -> Iterable[Child]:
  board = chess.Board(node.fen)
  for uci, p_prior in uci_priors.items():
    move = chess.Move.from_uci(uci)
    san = unchecked_san(board, move)
    logp_prior = np.log(p_prior)
    captured = captured_piece(board, move) if 'x' in san else None
    logp_ocr = lp(san, captured)
    if logp_ocr < logp_ocr_min:
      continue
    logp = agg_lp(logp_ocr, logp_prior)
    if logp >= logp_min:
      yield Child(fen=fen_after(move, board), logp=logp, san=san, parent=node)


from typing import Protocol, Sequence, AsyncIterable, Unpack, NotRequired, TypeAlias, Iterable
from haskellian import iter as I
from functools import partial
import chess
import lcz
from ..util import logprobs as logps
from .succs import AggregateLogps, SuccParams, Logprob, successors, Node

class UCIPrior(Protocol):
  """Evaluates a batch of `fens` into `UCI -> Probability` mappings"""
  def __call__(self, fens: Sequence[str], /) -> list[dict[str, float]]:
    ...

Beam: TypeAlias = Sequence[Node]

class StepParams(SuccParams):
  uci_prior: NotRequired[UCIPrior]
  agg_logp: NotRequired[AggregateLogps]

class FENlessSearchParams(StepParams):
  uci_prior: NotRequired[UCIPrior]
  agg_logp: NotRequired[AggregateLogps]
  beam_width: NotRequired[int]

class SearchParams(FENlessSearchParams):
  fen: NotRequired[str|None]

def step(
  beam: Sequence[Node], logprob: Logprob, *,
  uci_prior: UCIPrior = lcz.eval,
  agg_logp: AggregateLogps = partial(logps.weighted_geo_mean, a=10, b=1),
  **p: Unpack[SuccParams]
) -> Sequence[Node]:
  priors = uci_prior([n.fen for n in beam])
  succs = [
    successors(node, prior, logprob, agg_logp, **p)
    for node, prior in zip(beam, priors)
  ]
  return I.flatten(succs).sorted(key=lambda x: x.sum_logp, reverse=True)

async def search_async(
  logprobs: AsyncIterable[Logprob], *,
  fen: str | None = None,
  beam_width: int | None = None,
  **p: Unpack[StepParams]
) -> AsyncIterable[Beam]:
  """Beam search across the forest of moves stemming from `start_fens`
  - `logprobs[ply](san, piece)`: (OCR) log-probability of `san` (which captures `piece`) at `ply`
  - `uci_prior(fens)`: batched prior distribution of legal moves (defaults to using `MaiaChess` with `Leela Chess Zero`)
  - `agg_logp(logp, logq)`: aggregation function of the OCR and prior log-probabilities. Defaults to a weighted geometric average giving the OCR probabilities 10x the importance. I.e. `(p^10 * q)^(1/11)` (but in log-space, ofc)
  - `fen`: defaults to the starting position
  - `beam_width`: defaults to `4`
  """
  beam: Beam = [Node(fen or chess.STARTING_FEN)]
  async for lp in logprobs:
    beam = step(beam, lp, **p)[:beam_width or 4]
    if beam == []:
      return
    yield beam

def search_sync(
  logprobs: Iterable[Logprob], *,
  fen: str | None = None,
  beam_width: int | None = None,
  **p: Unpack[StepParams]
) -> Iterable[Beam]:
  """Beam search across the forest of moves stemming from `start_fens`
  - `logprobs[ply](san, piece)`: (OCR) log-probability of `san` (which captures `piece`) at `ply`
  - `uci_prior(fens)`: batched prior distribution of legal moves (defaults to using `MaiaChess` with `Leela Chess Zero`)
  - `agg_logp(logp, logq)`: aggregation function of the OCR and prior log-probabilities. Defaults to a weighted geometric average giving the OCR probabilities 10x the importance. I.e. `(p^10 * q)^(1/11)` (but in log-space, ofc)
  - `fen`: defaults to the starting position
  - `beam_width`: defaults to `4`
  """
  beam: Beam = [Node(fen or chess.STARTING_FEN)]
  for lp in logprobs:
    beam = step(beam, lp, **p)[:beam_width or 4]
    if beam == []:
      return
    yield beam


from editdistance import distance as edit_dist
from itertools import zip_longest
import numpy as np
from pydantic import BaseModel
from chess_notation import representations, Language, MotionStyles, PawnCapture, PieceCapture, CapturedPiece
from ..beam import Logprob

def pseudo_logp(word: str, top_preds: list[tuple[str, float]]) -> float:
  """Approximates the log probability of `word` given a subset of the best outputs `top_preds` of a model.
  - `top_preds :: [(Pred, Logprob)]`: the top-k predictions of a model with probability distribution `P`
    - So, each `(pred, logp) in top_preds` holds `logp = log P(pred)`
  - If `word` is in `top_preds`, the real log-probability is returned
  - Otherwise, an approximation factor `alpha(p)` is computed as `1 - NED(word, p)` for every `p in preds`
    - `NED` is the Normalized Edit Distance (normalized over `len(word)`)
    - If `NED(word, p) <= 1`, the assigned probability is `P(p)*alpha(p)`
    - Otherwise, the probability is 0
  - The maximum across all computed probabilities is returned
  
  **Note: both computations and the returned value are indeed log-probabilities. The explanation talks normal probabilities for clarity**
  """
  return max(
    (np.log(1 - dist/len(word)) + logp
    for pred, logp in top_preds if (dist := edit_dist(pred, word)) < min(len(word), len(pred))),
    default=-float('inf')
  )
  

class Annotations(BaseModel):
  lang: Language | None = None
  pawn_capture: PawnCapture | None = None
  piece_capture: PieceCapture | None = None

  def motions(self) -> MotionStyles:
    styles = MotionStyles()
    if self.pawn_capture: styles.pawn_captures = [self.pawn_capture]
    if self.piece_capture: styles.piece_captures = [self.piece_capture]
    return styles
  
  def langs(self) -> list[Language] | None:
    if self.lang is not None:
      return [self.lang]
    
def max_pseudo_logp(san: str, captured_piece: CapturedPiece | None, top_preds: list[tuple[str, float]], ann: Annotations) -> float:
  """Max `pseudo_logp` across all possible representations of `san`"""
  reprs = representations(san, motions=ann.motions(), captured_piece=captured_piece, languages=ann.langs() or ['CA', 'EN'])
  return max(pseudo_logp(r, top_preds) for r in reprs)

def players_max_pseudo_logp(san: str, captured_piece: CapturedPiece | None, players_preds: list[list[tuple[str, float]]], annotations: list[Annotations] | None = None) -> float:
  """Max `max_pseudo_logp` across players"""
  annotations = annotations or [Annotations() for _ in players_preds]
  return max(
    max_pseudo_logp(san, captured_piece, preds, ann or Annotations())
    for preds, ann in zip_longest(players_preds, annotations)
      if preds is not None
  )

def logprob(players_preds: list[list[tuple[str, float]]], annotations: list[Annotations] | None = None) -> Logprob:
  """Curried version of `players_max_pseudo_logp`"""
  def _logprob(san: str, captured_piece: CapturedPiece | None):
    return players_max_pseudo_logp(san, captured_piece, players_preds, annotations)
  return _logprob
  
def weighted_geo_mean(logp: float, logq: float, a: float, b: float):
  """Weighted geometrical mean (`[p^a * q^b]^(1/(a+b))`) but in log-space"""
  return (a*logp + b*logq) / (a+b)


from collections import Counter

def agg_max(preds: list[tuple[str, float]], k: int | None = None) -> list[tuple[str, float]]:
    """Aggregates by maxing probabilities of equal predictions"""
    acc = Counter()
    for move, prob in preds:
        cur = acc[move]
        # Counter defaults to 0, but logprob's identity is log(0) = -inf vvvvv
        acc[move] = max(prob, float("-inf") if cur == 0 else cur) # type: ignore
    return acc.most_common(k) # type: ignore

def agg_union(preds: list[tuple[str, float]], k: int | None = None) -> list[tuple[str, float]]:
    """Aggregates by summing probabilities of equal predictions"""
    import numpy as np
    acc = Counter()
    for move, prob in preds:
        acc[move] += np.exp(prob)
    return [(m, np.log(p)) for m, p in acc.most_common(k)]


from typing import Literal
from dataclasses import dataclass
from .beam import Child

@dataclass
class ManualPred:
  san: str
  tag: Literal['manual'] = 'manual'

@dataclass
class AutoPred:
  san: str
  prob: float
  tag: Literal['predicted'] = 'predicted'

  @classmethod
  def of(cls, child: Child) -> 'AutoPred':
    import numpy as np
    return AutoPred(san=child.san, prob=np.exp(child.logp))

Pred = ManualPred | AutoPred


from typing import AsyncIterable, Sequence, Iterable
from functools import partial
import numpy as np
from haskellian import iter as I, asyn_iter as AI
from .succs import Node, Child

def reconstruct(node: Node, max_depth: int | None = None) -> list[Child]:
  """Reconstruct backwards from `node` (keeping log-probs as they are)
  - `max_depth`: limit of backward steps
  """
  match node:
    case Child():
      if max_depth is None or max_depth > 0:
        return reconstruct(node.parent, max_depth=max_depth and max_depth-1) + [node]
      else:
        return [node]
    case _:
      return []
    
def exp(preds_logps: list[tuple[str, float]]) -> list[tuple[str, float]]:
  """Exponentiate log-probabilities"""
  preds, logps = I.unzip(preds_logps)
  exps = np.exp(logps)
  return list(zip(preds, exps))

def convergence(beam: Sequence[Node], max_depth: int | None = None) -> int:
  """Convergence point: ply back to which all lines of the `beam` agree on the best move
  - `max_depth`: limit of backward steps (from the current beam's ply)
  """
  paths = map(partial(reconstruct, max_depth=max_depth), beam)
  preds = map(lambda path: [node.san for node in path], paths)
  plymajor_preds = I.transpose(preds)
  ply_uniq_preds = map(set, plymajor_preds)
  uniq_plys = I.take_while(lambda uniq_preds: len(uniq_preds) == 1, ply_uniq_preds)
  return uniq_plys.len()
  
def decode_step(beam: Sequence[Node], ply: int, last_converged: int) -> list[Child] | None:
  converged = convergence(beam, max_depth=ply-last_converged)
  if converged > 0:
    return reconstruct(beam[0], max_depth=ply-last_converged)[:converged]
  
async def decode_async(beams: AsyncIterable[Sequence[Node]]) -> AsyncIterable[list[Child]]:
  last_converged = 0
  beam = None
  async for ply, beam in AI.enumerate(beams):
    converged = convergence(beam, max_depth=ply-last_converged)
    if converged > 0:
      yield reconstruct(beam[0], max_depth=ply-last_converged)[:converged]
      last_converged += converged
  if beam is not None:
    yield reconstruct(beam[0], max_depth=max(ply-last_converged, 0))

def decode_sync(beams: Iterable[Sequence[Node]]) -> Iterable[list[Child]]:
  last_converged = 0
  beam = None
  for ply, beam in enumerate(beams):
    converged = convergence(beam, max_depth=ply-last_converged)
    if converged > 0:
      yield reconstruct(beam[0], max_depth=ply-last_converged)[:converged]
      last_converged += converged
  if beam is not None:
    yield reconstruct(beam[0], max_depth=max(ply-last_converged, 0))


from typing import Protocol, Awaitable, AsyncIterable, TypedDict, NotRequired, Unpack
from haskellian import promise as P, asyn_iter as AI, iter as I
from .logprobs import logprob, Annotations
from ..beam.succs import Logprob

class PredictFn(Protocol):
  """Batched predictions of plies `[from_ply, to_ply)` for all (1 or 2) players
  - Must return an array of shape `BATCH x PLAYERS x TOP_PREDS` of `(pred, logprob)` tuples
  """
  def __call__(self, from_ply: int, to_ply: int) -> Awaitable[list[list[list[tuple[str, float]]]]]:
    ...

class BatchedParams(TypedDict):
  start_ply: NotRequired[int]
  batch_size: NotRequired[int]

class PrefetchedParams(BatchedParams):
  prefetch: NotRequired[int]

@AI.lift
async def batched_logprobs(predict: PredictFn, annotations: list[Annotations] | None = None, *, start_ply: int = 0, batch_size: int = 8) -> AsyncIterable[list[Logprob]]:
  for i in I.range(start_ply, step=batch_size):
    preds = await P.wait(predict(i, i+batch_size))
    if preds == []:
      return
    yield [logprob(ps, annotations) for ps in preds]

def prefetched_logprobs(predict: PredictFn, annotations: list[Annotations] | None = None, *, prefetch: int = 2, **p: Unpack[BatchedParams]) -> AsyncIterable[Logprob]:
  """Async iterable of `Logprob`s by calling `predict`. Prefetches and batches requests"""
  return batched_logprobs(predict, annotations, **p) \
    .prefetch(prefetch) \
    .f(AI.flatten).value
  


from .succs import Logprob, AggregateLogps, SuccParams, Node, Child
from .decoding import reconstruct, exp, convergence, decode_async, decode_sync, decode_step
from .searching import search_async, search_sync, step, SearchParams, FENlessSearchParams, StepParams, Beam
from .predict import predict_async, predict_sync

__all__ = [
  'Logprob', 'AggregateLogps', 'SuccParams', 'Node', 'Child',
  'reconstruct', 'exp', 'convergence', 'decode_async', 'decode_sync', 'decode_step',
  'search_async', 'search_sync', 'step', 'SearchParams', 'FENlessSearchParams', 'StepParams', 'Beam',
  'predict_sync', 'predict_async'
]



from .aggregate import agg_max, agg_union
from .logprobs import logprob
from .predict import prefetched_logprobs, PredictFn, PrefetchedParams, Annotations

__all__ = [
  'agg_max', 'agg_union',
  'logprob',
  'prefetched_logprobs', 'PredictFn', 'PrefetchedParams', 'Annotations'
]


from typing import Iterable, Unpack, Sequence
from haskellian import Iter, iter as I
import chess
from chess_utils import position_idx
from . import Pred, ManualPred, AutoPred, Annotations, predict_sync, predict_sync_raw, FENlessSearchParams

@I.lift
def manual_predict(
  preds: list[list[list[tuple[str, float]]]],
  manual_ucis: dict[int, str] = {},
  annotations: list[Annotations] | None = None,
  fen: str | None = None,
  **p: Unpack[FENlessSearchParams]
) -> Iterable[Sequence[Pred]]:
  """Like `predict_sync` but must pass through moves specified in `manual_ucis`
  - `manual_ucis[ply] = move` indicates a move to make at index `ply`
  - If the move isn't legal, the search will stop
  """
  fen = fen or chess.STARTING_FEN
  idx = position_idx(fen)
  next_manual = Iter(manual_ucis).filter(lambda i: i >= idx).min()
  
  if next_manual is None:
    yield from predict_sync(preds, annotations, fen=fen, *p)

  elif next_manual == idx:
    board = chess.Board(fen)
    try:
      move = board.push_uci(manual_ucis[idx])
      fen = board.fen()
      board.pop()
      yield [ManualPred(san=board.san(move))]
      yield from manual_predict(preds, manual_ucis, annotations, fen=fen, **p)
    except chess.IllegalMoveError:
      ...
  
  else: # next_manual > idx
    child = None
    for child in predict_sync_raw(preds, annotations, fen=fen, **p).flatmap().take(next_manual-idx):
      yield [AutoPred.of(child)]
    if child and position_idx(child.fen) == next_manual:
      yield from manual_predict(preds, manual_ucis, annotations, fen=child.fen)


from typing import AsyncIterable, Iterable, Unpack
from haskellian import asyn_iter as AI, iter as I
from .succs import Logprob
from .searching import search_sync, search_async, SearchParams
from .decoding import decode_async, decode_sync

@AI.lift
def predict_async(logprobs: AsyncIterable[Logprob], **params: Unpack[SearchParams]):
  """Beam decoding across the forest of moves stemming from `start_fens`
  - Yields predictions as the beams converge (i.e. agree on a single move) or the search stops (because no legal moves have high enough probability)
    - Thus, a bigger `beam_width` can increase accuracy but also prediction time by more than a constant factor
    
  Params:
  - `logprobs[ply](san, piece)`: (OCR) log-probability of `san` (which captures `piece`) at `ply`
  - `uci_prior(fens)`: batched prior distribution of legal moves (defaults to using `MaiaChess` with `Leela Chess Zero`)
  - `agg_logp(logp, logq)`: aggregation function of the OCR and prior log-probabilities. Defaults to a weighted geometric average giving the OCR probabilities 10x the importance. I.e. `(p^10 * q)^(1/11)` (but in log-space, ofc)
  """
  beams = search_async(logprobs, **params)
  return decode_async(beams)

@I.lift
def predict_sync(logprobs: Iterable[Logprob], **params: Unpack[SearchParams]):
  """Beam decoding across the forest of moves stemming from `start_fens`
  - Yields predictions as the beams converge (i.e. agree on a single move) or the search stops (because no legal moves have high enough probability)
    - Thus, a bigger `beam_width` can increase accuracy but also prediction time by more than a constant factor
    
  Params:
  - `logprobs[ply](san, piece)`: (OCR) log-probability of `san` (which captures `piece`) at `ply`
  - `uci_prior(fens)`: batched prior distribution of legal moves (defaults to using `MaiaChess` with `Leela Chess Zero`)
  - `agg_logp(logp, logq)`: aggregation function of the OCR and prior log-probabilities. Defaults to a weighted geometric average giving the OCR probabilities 10x the importance. I.e. `(p^10 * q)^(1/11)` (but in log-space, ofc)
  """
  beams = search_sync(logprobs, **params)
  return decode_sync(beams)


from .util import Annotations
from .beam import SearchParams, Child, FENlessSearchParams
from . import beam
from .types import Pred, AutoPred, ManualPred
from .main import predict, predict_async, predict_sync, predict_sync_raw
from .manual import manual_predict

__all__ = [
  'predict', 'predict_async', 'predict_sync', 'predict_sync_raw', 'Annotations', 'SearchParams', 'beam', 'Child',
  'Pred', 'AutoPred', 'ManualPred', 'FENlessSearchParams', 'manual_predict'
]